# coding=UTF-8
#
#	Copyright 2013-2014, Marten de Vries
#
#	Licensed under the Apache License, Version 2.0 (the "License");
#	you may not use this file except in compliance with the License.
#	You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
#	Unless required by applicable law or agreed to in writing, software
#	distributed under the License is distributed on an "AS IS" BASIS,
#	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#	See the License for the specific language governing permissions and
#	limitations under the License.

import contextlib
import pouchdb
import inspect
import textwrap
import sys

@contextlib.contextmanager
def suppress(*e):
	try:
		yield
	except e:
		pass

class PromiseStub(object):
	def __init__(self, *args, **kwargs):
		super(PromiseStub, self).__init__(*args, **kwargs)

		self._successHandlers = set()
		self._failureHandlers = set()
		self.done = False

	def then(self, h):
		self._successHandlers.add(h)

	def catch(self, h):
		self._failureHandlers.add(h)

	def success(self, resp):
		for h in self._successHandlers:
			h(resp)
		self.done = True

	def failure(self, err):
		"""Assumes ``err`` to be sys.exc_info()'s output."""

		if not self._failureHandlers:
			self.done = True
			raise err[1], None, err[2]
		err[1].traceback = err[2]
		for h in self._failureHandlers:
			h(err[1])
		self.done = True

class JSON(object):
	def __init__(self, data):
		self.data = data

class Blob(object):
	def __init__(self, data):
		self.data = data

class JSError(Exception):
	def __init__(self, messages, *args, **kwargs):
		message = "\n".join(messages)

		super(JSError, self).__init__(message, *args, **kwargs)

class AttrAccessDict(dict):
	def __getattr__(self, key):
		try:
			return super(AttrAccessDict, self).__getattr__(key)
		except AttributeError:
			try:
				return self[key]
			except KeyError, e:
				raise AttributeError(e)

def getPromiseResult(ctx, promise):
	returned = {}
	def success(resp):
		returned["resp"] = resp
	def fail(err):
		returned["err"] = err

	promise.then(success, fail)
	ctx.waitUntilCalled([success, fail])

	if returned.get("err"):
		raise pouchdb.PouchDBError(returned["err"])
	return returned["resp"]

@contextlib.contextmanager
def gettingCallbackResult(ctx):
	returning = {}
	def callback(err=None, resp=None):
		returning["err"] = err
		returning["resp"] = resp
	info = {"callback": callback}
	yield info
	ctx.waitUntilCalled(callback)
	if returning.get("err"):
		raise pouchdb.PouchDBError(returning["err"])
	info["resp"] = returning["resp"]

def filterNone(args):
	return [a for a in args if a is not None]

def jsonAllowed(value):
	if isinstance(value, basestring) and (value.startswith("{") or value.startswith("[")):
		return JSON(value)
	return value

def aliasFor(method):
	methodName = method.__name__
	def alias(self, *args, **kwargs):
		return getattr(self, methodName)(*args, **kwargs)
	alias.__doc__ = "Alias for :meth:`%s`." % methodName
	return alias

@contextlib.contextmanager
def jsToPouchError():
	try:
		yield
	except JSError, e:
		raise pouchdb.PouchDBError(e)

@contextlib.contextmanager
def jsTypeErrorToAttributeError():
	try:
		yield
	except JSError, e:
		if "TypeError" in str(e):
			raise AttributeError(e)
		raise #pragma: no cover

class _AutoImplementer(object):
	def __call__(self,  type):
		def wrapper(cls):
			for methodName in cls.__abstractmethods__:
				abstractMethod = getattr(cls, methodName)
				method = self._buildMethod(type, cls, abstractMethod)
				setattr(cls, methodName, method)
			cls.__abstractmethods__ = frozenset()

			return cls
		return wrapper

	def _buildMethod(self, type, cls, abstractMethod):
		data = {}
		data["name"] = abstractMethod.__name__
		data["jsName"] = getattr(abstractMethod, "jsName", data["name"])

		wrapJSError = getattr(abstractMethod, "wrapJSError", False)
		data["errorWrapper"] = ""
		if wrapJSError:
			data["errorWrapper"] = "with jsToPouchError():"

		wrapTypeError = getattr(abstractMethod, "wrapTypeError", False)
		if wrapTypeError:
			data["errorWrapper"] = "with jsTypeErrorToAttributeError():"
		if data["errorWrapper"]:
			data["errorWrapper"] += "\n" + 5 * "\t"

		data["funcName"] = "_asyncCall" if type == "async" else "_syncCall"
		data.update(self._buildArgs(type, abstractMethod))

		data["addKwargs"] = ""
		if data["kwargs"]:
			#filter out the keyword argument if it's just '{}'.
			data["addKwargs"] += "if not {kwargs}: args = [a for a in args if a is not {kwargs}]".format(**data)

		code = textwrap.dedent("""
			def {name}{formattedArgs}:
				args = [{funcArgs}]
				{addKwargs}
				{errorWrapper}return self.{funcName}("{jsName}", *args)
		""".format(**data))

		filename = "<generated source of '%s.%s'>" % (cls.__name__, data["name"])
		try:
			codeObj = compile(code, filename, "exec")
		except SyntaxError:#pragma: no cover
			#Runs only when a mistake exists in Python-PouchDB's code.
			print >> sys.stderr, code
			raise
		namespace = {
			"jsonAllowed": jsonAllowed,
			"jsToPouchError": jsToPouchError,
			"jsTypeErrorToAttributeError": jsTypeErrorToAttributeError,
		}
		exec(codeObj, namespace)
		return namespace[data["name"]]

	def _buildArgs(self, type, abstractMethod):
		args, varargs, keywords, defaults = inspect.getargspec(abstractMethod)
		defaults = list(defaults or ())

		jsonArgs = set(getattr(abstractMethod, "allowsJSON", []))

		funcArgs = args[1:]
		for i, arg in enumerate(funcArgs):
			if arg in jsonArgs:
				funcArgs[i] = "jsonAllowed(%s)" % arg
				jsonArgs.discard(arg)
		if jsonArgs:#pragma: no cover
			#Runs only when a mistake exists in Python-PouchDB's code.
			raise ValueError("Unexisting arguments: %s" % jsonArgs)

		if keywords:
			funcArgs.append(keywords)

		hasNoCallback = getattr(abstractMethod, "hasNoCallback", False)
		if type == "async" and not hasNoCallback:
			args.append("callback")
			funcArgs.append("callback")
			defaults.append(None)

		return {
			"formattedArgs": inspect.formatargspec(args, varargs, keywords, defaults),
			"funcArgs": ", ".join(funcArgs),
			"kwargs": keywords,
		}

autoImplementMissing = _AutoImplementer()

def allowsJSON(*args):
	def wrapper(func):
		func.allowsJSON = set(args)
		return func
	return wrapper

def needsJSErrorWrapper(func):
	func.wrapJSError = True
	return func

def needsJSTypeErrorWrapper(func):
	func.wrapTypeError = True
	return func

def noCallback(func):
	func.hasNoCallback = True
	return func

def withJSName(name):
	def wrapper(func):
		func.jsName = name
		return func
	return wrapper

def handlePromise(g, promise, cb):
	def onSuccess(resp):
		with cbOnException(cb):
			handlePromise(g, g.send(resp), cb)
	def onError(err):
		exc = pouchdb.PouchDBError(err)
		with cbOnException(cb):
			handlePromise(g, g.throw(exc), cb)
	promise.then(onSuccess)
	promise.catch(onError)

@contextlib.contextmanager
def cbOnException(cb):
	try:
		yield
	except Exception, e:
		cb(sys.exc_info())
